@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2009 Sergey Poznyakoff
+   Copyright (C) 2009-2019 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,11 +28,11 @@
 
 #define IFLD_MIN_COUNT   6	/* Minimum number of fields in entry */
 
-/* FIXME: Duplicated in grex-lex.l */
+/* FIXME: Copied from grecs/src/tree.c */
 static void
-listel_dispose(const void *el)
+listel_dispose(void *el)
 {
-  free((void*)el);
+  free(el);
 }
 
 #define TCPMUX_PREFIX_STR "tcpmux/"
@@ -64,13 +64,13 @@ mktag (const char *address, const char *service)
   
   if (address)
     {
-      str = xmalloc (strlen (address) + 1 + strlen (service) + 1);
+      str = grecs_malloc (strlen (address) + 1 + strlen (service) + 1);
       strcpy (str, address);
       strcat (str, ":");
       strcat (str, service);
     }
   else
-    str = xstrdup (service);
+    str = grecs_strdup (service);
   return str;
 }
 
@@ -81,9 +81,10 @@ inetd_conf_file (const char *file)
   FILE *fp;
   size_t size = 0;
   char *buf = NULL;
-  size_t line_no = 0;
+  unsigned long line_no = 0;
   struct wordsplit ws;
   char *dfl_address = NULL;
+  int wsflags;
   
   fp = fopen (file, "r");
   if (!fp)
@@ -94,6 +95,7 @@ inetd_conf_file (const char *file)
       return 1;
     }
 
+  wsflags = WRDSF_NOVAR | WRDSF_NOCMD | WRDSF_WS | WRDSF_SQUEEZE_DELIMS;
   while (getline (&buf, &size, fp) >= 0)
     {
       char *p;      
@@ -111,9 +113,6 @@ inetd_conf_file (const char *file)
       struct inetd_builtin *builtin;
       char *tag;
       
-      if (line_no)
-	wordsplit_free (&ws);
-      
       line_no++;
       for (p = buf; *p && c_isblank (*p); p++)
 	;
@@ -121,14 +120,12 @@ inetd_conf_file (const char *file)
       if (!p || *p == '\n' || *p == '#')
 	continue;
 
-      if (wordsplit (p, &ws,
-		     WRDSF_NOVAR | WRDSF_NOCMD | WRDSF_WS |
-		     WRDSF_SQUEEZE_DELIMS))
+      if (wordsplit (p, &ws, wsflags))
 	{
 	  logmsg (LOG_ERR, "wordsplit: %s", strerror (errno));
 	  continue;
 	}
-
+      wsflags |= WRDSF_REUSE;
       if (ws.ws_wordc == 1)
 	{
 	  size_t len = strlen (ws.ws_wordv[IFLD_SERVICE]);
@@ -140,7 +137,7 @@ inetd_conf_file (const char *file)
 		dfl_address = NULL;
 	      else
 		{
-		  dfl_address = xmalloc (len);
+		  dfl_address = grecs_malloc (len);
 		  memcpy (dfl_address, ws.ws_wordv[IFLD_SERVICE], len-1);
 		  dfl_address[len-1] = 0;
 		}
@@ -196,19 +193,25 @@ inetd_conf_file (const char *file)
 	}
       else
 	{
+	  char *s = NULL;
+	  size_t l = 0;
+	  int rc;
+	  
 	  /* Create URL from protocol and service fields. */
-	  str = xasprintf ("inet+%s://%s:%s",
-			   ws.ws_wordv[IFLD_PROTOCOL],
-			   address ? address : "0.0.0.0",
-			   service);
-	  if (pies_url_create (&url, str))
+	  if (grecs_asprintf (&s, &l, "inet+%s://%s:%s",
+			      ws.ws_wordv[IFLD_PROTOCOL],
+			      address ? address : "0.0.0.0",
+			      service))
+	    grecs_alloc_die ();
+	  rc = pies_url_create (&url, s);
+	  free (s);
+	  if (rc)
 	    {
 	      /* FIXME: Better error message */
 	      logmsg (LOG_ERR, "%s:%lu: %s",
 		      file, line_no, _("invalid socket address"));
 	      continue;
 	    }
-	  free (str);
 	}
       
       /* Parse wait/nowait field */
@@ -277,42 +280,35 @@ inetd_conf_file (const char *file)
 	comp->flags = flags;
       if (ISCF_TCPMUX (comp->flags))
 	comp->tcpmux = mktag (address, "tcpmux");
-      comp->service = xstrdup (service);
-      comp->privs.user = xstrdup (user); /* FIXME: memory leak */
+      comp->service = grecs_strdup (service);
+      comp->privs.user = grecs_strdup (user);
       if (group)
 	{
-	  comp->privs.groups =
-	    gl_list_create_empty (&gl_linked_list_implementation,
-				  NULL,
-				  NULL,
-				  listel_dispose,
-				  true);
-	  gl_list_add_last (comp->privs.groups, xstrdup (group));
+	  comp->privs.groups = grecs_list_create ();
+	  comp->privs.groups->free_entry = listel_dispose;
+	  grecs_list_append (comp->privs.groups, grecs_strdup (group));
 	}
 
-      comp->program = xstrdup (ws.ws_wordv[IFLD_SERVER_PATH]);
+      comp->program = grecs_strdup (ws.ws_wordv[IFLD_SERVER_PATH]);
 
       if (ws.ws_wordc > IFLD_MIN_COUNT)
 	{
 	  size_t i, j;
 	  
 	  comp->argc = ws.ws_wordc - IFLD_MIN_COUNT;
-	  comp->argv = xcalloc (comp->argc + 1, sizeof (comp->argv[0]));
+	  comp->argv = grecs_calloc (comp->argc + 1, sizeof (comp->argv[0]));
 	  for (i = IFLD_SERVER_ARGS, j = 0; i < ws.ws_wordc; i++, j++)
-	    comp->argv[j] = xstrdup (ws.ws_wordv[i]);
+	    comp->argv[j] = grecs_strdup (ws.ws_wordv[i]);
 	}
       else
 	{
 	  comp->argc = 1;
-	  comp->argv = xcalloc (comp->argc + 1, sizeof (comp->argv[0]));
-	  comp->argv[0] = xstrdup (comp->program);
+	  comp->argv = grecs_calloc (comp->argc + 1, sizeof (comp->argv[0]));
+	  comp->argv[0] = grecs_strdup (comp->program);
 	}
-
-      if (progman_lookup_component (comp->tag) == NULL)
-	register_prog (comp);
     }
   
-  if (line_no)
+  if (wsflags & WRDSF_REUSE)
     wordsplit_free (&ws);
   free (dfl_address);
   free (buf);
@@ -348,7 +344,7 @@ inetd_conf_dir (const char *name)
   if (name[namelen-1] != '/')
     namebufsize++;
   namebufsize += NAME_INIT_ALLOC;
-  namebuf = xmalloc (namebufsize);
+  namebuf = grecs_malloc (namebufsize);
   memcpy (namebuf, name, namelen);
   if (name[namelen-1] != '/')
     namebuf[namelen++] = 0;
@@ -359,8 +355,10 @@ inetd_conf_dir (const char *name)
       
       if (stat (ent->d_name, &st))
 	{
-	  logmsg (LOG_ERR, _("cannot stat %s/%s: %s"),
-		  name, ent->d_name, strerror (errno));
+	  int ec = errno;
+	  char *name = mkfilename (name, ent->d_name, NULL);
+	  logfuncall ("stat", name, ec);
+	  free (name);
 	  errs |= 1;
 	}
       else if (S_ISREG (st.st_mode) || S_ISLNK (st.st_mode))
@@ -369,7 +367,7 @@ inetd_conf_dir (const char *name)
 	  if (namelen + len >= namebufsize)
 	    {
 	      namebufsize = namelen + len + 1;
-	      namebuf = xrealloc (namebuf, namebufsize);
+	      namebuf = grecs_realloc (namebuf, namebufsize);
 	    }
 	  strcpy (namebuf + namelen, ent->d_name);
 	  errs |= inetd_conf_file (namebuf);
@@ -381,13 +379,13 @@ inetd_conf_dir (const char *name)
 }
 
 int
-inetd_parse_conf (const char *file)
+inetd_config_parse (const char *file)
 {
   struct stat st;
 
   if (stat (file, &st))
     {
-      logmsg (LOG_ERR, _("cannot stat %s: %s"), file, strerror (errno));
+      logfuncall ("stat", file, errno);
       return 1;
     }
   if (S_ISDIR (st.st_mode))
